/* 

File: nesta.cpp

Copyright (c) 2010 Department of Information Engineering, University of Padova, Italy
Contributors: Davide Zordan, Feb. 2010
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.
    * The name of the author may not be used to endorse or promote
      products derived from this software without specific prior
      written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "nesta.h"

void PrintMatrix(const Matrix& X) {
	cout << "\nMatrix type: " << X.Type().Value() << " (";
	cout << X.Nrows() << ", ";
	cout << X.Ncols() << ")\n\n";
	if (X.IsZero()) {
		cout << "All elements are zero\n" << flush;
		return;
	}
	int nr = X.Nrows();
	int nc = X.Ncols();
	for (int i = 1; i <= nr; i++) {
		for (int j = 1; j <= nc; j++)
			cout << X(i, j) << "\t";
		cout << "\n";
	}
	cout << flush;
}

double myabs(double x) {
	if (x >= 0)
		return x;
	else
		return (-x);
}

double norm(const ColumnVector& X) {
	return (sqrt(X.SumSquare()));
}

double media(double array[]) {
	int num = 0;
	double somma = 0;

	for (int i = 0; i < 10; i++)
		if (array[i] != 0) {
			num += 1;
			somma += array[i];
		}
	if (somma == 0) {
		return 1e-100;
	}
	return (somma / num);
}

ColumnVector Nesterov(const Matrix& A, const ColumnVector& y) {
	float mu = 0.01;
	float delta = 1e-5;
	float epsilon = 1e-3;

	double maxiter = 10000;
	bool verbose = 0;

	int N = A.Ncols();     // number of elements in x

	ColumnVector x_k(N);   // initialize x_0  (x_0=A'b)
	x_k << A.t() * y;
	if (verbose)
		PrintMatrix(x_k);

	ColumnVector w_k(N);
	w_k = 0.0;               // initialize w_k, y_k, z_k
	ColumnVector y_k(N);
	y_k = x_k;
	ColumnVector z_k(N);
	z_k = x_k;

	double A_k = 0;
	double L_mu = 1 / mu;

	ColumnVector Atb(N);
	Atb << x_k;

	ColumnVector df(N);         // grad_f_mu
	ColumnVector val_f_mu(1);   // f_mu(x_k)
	ColumnVector q(N);          // q in eq.3.7 paper NESTA
	ColumnVector Ayk(N);        // not used
	ColumnVector Azk(N);
	double lambda = 0;
	double gamma = 0;
	double alpha_k = 0;
	double tau_k = 0;
	double f_mean[10];
	double delta_f_mu = 0;
	int niter = 0;
	for (int k = 0; k < maxiter; k++) {
		niter++;
		// evaluate f_mu(x_k), grad_f_mu //
		for (int j = 1; j <= N; j++) {
			if (myabs(x_k(j)) > mu) {
				df(j) = x_k(j) / myabs(x_k(j));
			} else {
				df(j) = x_k(j) / mu;
			}
		}
		val_f_mu << (df.t() * x_k) - ((mu / 2) * (pow(norm(df), 2)));

		if (verbose)
			PrintMatrix(df);
		if (verbose)
			PrintMatrix(val_f_mu);

		// updating y_k //
		q << x_k - (1 / L_mu) * df;
		if (verbose)
			PrintMatrix(q);

		if (epsilon > 0) {
			lambda = max(0.0, L_mu * norm(y - (A * q)) / epsilon - 1);
			//cout << lambda;
			gamma = lambda / (lambda + L_mu);
			//cout << gamma;
			y_k << lambda / L_mu * (1 - gamma) * Atb + q - gamma * A.t() * A * q;
			//PrintMatrix(y_k);
			Ayk << lambda / L_mu * (1 - gamma) * y + A * q - gamma * A * q;
			//PrintMatrix(Ayk);
		} else {
			y_k = A.t() * A * Atb + q - A.t() * A * q;
			//PrintMatrix(y_k);
			Ayk = y;
			//PrintMatrix(Ayk);
		}

		// stopping criterion //

		delta_f_mu = myabs(val_f_mu(1) - media(f_mean)) / media(f_mean);

		if (delta_f_mu <= delta) {
			break;
		}

		for (int j = 2; j <= 10; j++) {
			f_mean[11 - j] = f_mean[10 - j];
		}
		f_mean[0] = val_f_mu(1);   // new value

		// updating z_k
		alpha_k = 0.5 * (k + 1); // see the paper NESTA
		A_k += alpha_k;
		tau_k = 2 / (k + 3);

		w_k << alpha_k * df + w_k;
		q << Atb - (1 / L_mu) * w_k;

		if (epsilon > 0) {
			lambda = max(0.0, L_mu * norm(y - (A * q)) / epsilon - 1);
			//cout << lambda;
			gamma = lambda / (lambda + L_mu);
			//cout << gamma;
			z_k << lambda / L_mu * (1 - gamma) * Atb + q - gamma * A.t() * A * q;
			//PrintMatrix(y_k);
			Azk << lambda / L_mu * (1 - gamma) * y + A * q - gamma * A * q;
			//PrintMatrix(Ayk);
		} else {
			z_k = A.t() * A * Atb + q - A.t() * A * q;
			//PrintMatrix(y_k);
			Azk = y;
			//PrintMatrix(Ayk);
		}

		x_k << tau_k * z_k + (1 - tau_k) * y_k;
	}
	return x_k;
}

ColumnVector mean(const Matrix& X) {
	int N = X.Nrows();
	int M = X.Ncols();
	ColumnVector m(N);
	m = 0;
	for (int i = 1; i <= M; i++) {
		m = m + X.Column(i);
	}
	m = m / M;
	return m;
}

SymmetricMatrix cov(Matrix X) {
	int N = X.Nrows();
	int M = X.Ncols();
	ColumnVector R(N);
	ColumnVector m(N);
	m = mean(X);
	SymmetricMatrix Sigma(N);
	Sigma = 0;
	for (int i = 1; i <= M; i++) {
		R << X.Column(i);
		Sigma << Sigma + (R - m) * ((R - m).t());
	}
	return Sigma;
}

void PCA_Transform_Matrix(Matrix TS, Matrix& Psi, ColumnVector& m) {
	int N = TS.Nrows();
	SymmetricMatrix S(N);
	m = mean(TS);
	S = cov(TS);
//	PrintMatrix(S);
	Matrix V;
	DiagonalMatrix D;

	EigenValues(S, D, V);

	for (int i = 1; i <= N; i++) {
		Psi.Column(i) = V.Column(N - i + 1);
	}
}
